package mixins

import (
	dms3ld "gitlab.dms3.io/ld/go-ld-prime"
)

// Bytes can be embedded in a struct to provide all the methods that
// have fixed output for any int-kinded nodes.
// (Mostly this includes all the methods which simply return ErrWrongKind.)
// Other methods will still need to be implemented to finish conforming to Node.
//
// To conserve memory and get a TypeName in errors without embedding,
// write methods on your type with a body that simply initializes this struct
// and immediately uses the relevant method;
// this is more verbose in source, but compiles to a tighter result:
// in memory, there's no embed; and in runtime, the calls will be inlined
// and thus have no cost in execution time.
type Bytes struct {
	TypeName string
}

func (Bytes) Kind() dms3ld.Kind {
	return dms3ld.Kind_Bytes
}
func (x Bytes) LookupByString(string) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByString", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) LookupByNode(key dms3ld.Node) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByNode", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) LookupByIndex(idx int64) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByIndex", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) LookupBySegment(dms3ld.PathSegment) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupBySegment", AppropriateKind: dms3ld.KindSet_Recursive, ActualKind: dms3ld.Kind_Bytes}
}
func (Bytes) MapIterator() dms3ld.MapIterator {
	return nil
}
func (Bytes) ListIterator() dms3ld.ListIterator {
	return nil
}
func (Bytes) Length() int64 {
	return -1
}
func (Bytes) IsAbsent() bool {
	return false
}
func (Bytes) IsNull() bool {
	return false
}
func (x Bytes) AsBool() (bool, error) {
	return false, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) AsInt() (int64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) AsFloat() (float64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) AsString() (string, error) {
	return "", dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsString", AppropriateKind: dms3ld.KindSet_JustString, ActualKind: dms3ld.Kind_Bytes}
}
func (x Bytes) AsLink() (dms3ld.Link, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_Bytes}
}

// BytesAssembler has similar purpose as Bytes, but for (you guessed it)
// the NodeAssembler interface rather than the Node interface.
type BytesAssembler struct {
	TypeName string
}

func (x BytesAssembler) BeginMap(sizeHint int64) (dms3ld.MapAssembler, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "BeginMap", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) BeginList(sizeHint int64) (dms3ld.ListAssembler, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "BeginList", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignNull() error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignNull", AppropriateKind: dms3ld.KindSet_JustNull, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignBool(bool) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignInt(int64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignFloat(float64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignString(string) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignString", AppropriateKind: dms3ld.KindSet_JustString, ActualKind: dms3ld.Kind_Bytes}
}
func (x BytesAssembler) AssignLink(dms3ld.Link) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_Bytes}
}
