package mixins

import (
	dms3ld "gitlab.dms3.io/ld/go-ld-prime"
)

// Map can be embedded in a struct to provide all the methods that
// have fixed output for any map-kinded nodes.
// (Mostly this includes all the methods which simply return ErrWrongKind.)
// Other methods will still need to be implemented to finish conforming to Node.
//
// To conserve memory and get a TypeName in errors without embedding,
// write methods on your type with a body that simply initializes this struct
// and immediately uses the relevant method;
// this is more verbose in source, but compiles to a tighter result:
// in memory, there's no embed; and in runtime, the calls will be inlined
// and thus have no cost in execution time.
type Map struct {
	TypeName string
}

func (Map) Kind() dms3ld.Kind {
	return dms3ld.Kind_Map
}
func (x Map) LookupByIndex(idx int64) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByIndex", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_Map}
}
func (Map) ListIterator() dms3ld.ListIterator {
	return nil
}
func (Map) IsAbsent() bool {
	return false
}
func (Map) IsNull() bool {
	return false
}
func (x Map) AsBool() (bool, error) {
	return false, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_Map}
}
func (x Map) AsInt() (int64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_Map}
}
func (x Map) AsFloat() (float64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_Map}
}
func (x Map) AsString() (string, error) {
	return "", dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsString", AppropriateKind: dms3ld.KindSet_JustString, ActualKind: dms3ld.Kind_Map}
}
func (x Map) AsBytes() ([]byte, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsBytes", AppropriateKind: dms3ld.KindSet_JustBytes, ActualKind: dms3ld.Kind_Map}
}
func (x Map) AsLink() (dms3ld.Link, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_Map}
}

// MapAssembler has similar purpose as Map, but for (you guessed it)
// the NodeAssembler interface rather than the Node interface.
type MapAssembler struct {
	TypeName string
}

func (x MapAssembler) BeginList(sizeHint int64) (dms3ld.ListAssembler, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "BeginList", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignNull() error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignNull", AppropriateKind: dms3ld.KindSet_JustNull, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignBool(bool) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignInt(int64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignFloat(float64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignString(string) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignString", AppropriateKind: dms3ld.KindSet_JustString, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignBytes([]byte) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignBytes", AppropriateKind: dms3ld.KindSet_JustBytes, ActualKind: dms3ld.Kind_Map}
}
func (x MapAssembler) AssignLink(dms3ld.Link) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_Map}
}
