package mixins

import (
	dms3ld "gitlab.dms3.io/ld/go-ld-prime"
)

// String can be embedded in a struct to provide all the methods that
// have fixed output for any string-kinded nodes.
// (Mostly this includes all the methods which simply return ErrWrongKind.)
// Other methods will still need to be implemented to finish conforming to Node.
//
// To conserve memory and get a TypeName in errors without embedding,
// write methods on your type with a body that simply initializes this struct
// and immediately uses the relevant method;
// this is more verbose in source, but compiles to a tighter result:
// in memory, there's no embed; and in runtime, the calls will be inlined
// and thus have no cost in execution time.
type String struct {
	TypeName string
}

func (String) Kind() dms3ld.Kind {
	return dms3ld.Kind_String
}
func (x String) LookupByString(string) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByString", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_String}
}
func (x String) LookupByNode(key dms3ld.Node) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByNode", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_String}
}
func (x String) LookupByIndex(idx int64) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupByIndex", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_String}
}
func (x String) LookupBySegment(dms3ld.PathSegment) (dms3ld.Node, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "LookupBySegment", AppropriateKind: dms3ld.KindSet_Recursive, ActualKind: dms3ld.Kind_String}
}
func (String) MapIterator() dms3ld.MapIterator {
	return nil
}
func (String) ListIterator() dms3ld.ListIterator {
	return nil
}
func (String) Length() int64 {
	return -1
}
func (String) IsAbsent() bool {
	return false
}
func (String) IsNull() bool {
	return false
}
func (x String) AsBool() (bool, error) {
	return false, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_String}
}
func (x String) AsInt() (int64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_String}
}
func (x String) AsFloat() (float64, error) {
	return 0, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_String}
}
func (x String) AsBytes() ([]byte, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsBytes", AppropriateKind: dms3ld.KindSet_JustBytes, ActualKind: dms3ld.Kind_String}
}
func (x String) AsLink() (dms3ld.Link, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AsLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_String}
}

// StringAssembler has similar purpose as String, but for (you guessed it)
// the NodeAssembler interface rather than the Node interface.
type StringAssembler struct {
	TypeName string
}

func (x StringAssembler) BeginMap(sizeHint int64) (dms3ld.MapAssembler, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "BeginMap", AppropriateKind: dms3ld.KindSet_JustMap, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) BeginList(sizeHint int64) (dms3ld.ListAssembler, error) {
	return nil, dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "BeginList", AppropriateKind: dms3ld.KindSet_JustList, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignNull() error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignNull", AppropriateKind: dms3ld.KindSet_JustNull, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignBool(bool) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignBool", AppropriateKind: dms3ld.KindSet_JustBool, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignInt(int64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignInt", AppropriateKind: dms3ld.KindSet_JustInt, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignFloat(float64) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignFloat", AppropriateKind: dms3ld.KindSet_JustFloat, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignBytes([]byte) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignBytes", AppropriateKind: dms3ld.KindSet_JustBytes, ActualKind: dms3ld.Kind_String}
}
func (x StringAssembler) AssignLink(dms3ld.Link) error {
	return dms3ld.ErrWrongKind{TypeName: x.TypeName, MethodName: "AssignLink", AppropriateKind: dms3ld.KindSet_JustLink, ActualKind: dms3ld.Kind_String}
}
